#include "uploader.h"

Uploader::Uploader()
{
    // "applications must ensure that create() is only called on the main (GUI) thread"
    m_surface.create();

    m_copy_source = QSharedPointer<QVector<copy>>(new QVector<copy>);
    m_draw_textures = QSharedPointer<QVector<QOpenGLTexture*>>(new QVector<QOpenGLTexture*>);

    // A single rectangle takes 20 float values to describe (4 vertices * (2 coordinate + 3 color))
    m_fill_data.reserve(100);
    m_draw_data.reserve(100);
    m_copy_data.reserve(5);
    m_copy_source->reserve(5);
    m_draw_textures->reserve(5);
}

void Uploader::create_context(QOpenGLContext *context)
{
    m_uploader_context = new QOpenGLContext();
    m_uploader_context->setShareContext(context);
    m_uploader_context->create();
    m_uploader_context->makeCurrent(&m_surface);
    m_uploader_context->functions()->initializeOpenGLFunctions();

    m_fill_buffer.create();
    m_copy_buffer.create();
    m_draw_buffer.create();
}

void Uploader::startedUpdate()
{
    // Clean up the data from the previous round
    m_fill_data.clear();
    m_fill_count = 0;

    for(unsigned int i = 0; i < m_copy_count; i++) {
        m_copy_source->at(i).texture->destroy();
        delete m_copy_source->at(i).texture;
    }
    m_copy_source->clear();
    m_copy_data.clear();
    m_copy_count = 0;

    for(unsigned int i = 0; i < m_draw_count; i++) {
        m_draw_textures->at(i)->destroy();
        delete m_draw_textures->at(i);
    }
    m_draw_textures->clear();
    m_draw_data.clear();
    m_draw_count = 0;
}

void Uploader::gotCopyRect(const int src_x, const int src_y, const int width, const int height, const int dest_x, const int dest_y)
{
    if(m_framebufferObject != NULL && !m_framebufferObject->isBound())
        m_framebufferObject->bind();

    QOpenGLTexture* texture = new QOpenGLTexture(QOpenGLTexture::Target2D);
    texture->setAutoMipMapGenerationEnabled(false);
    texture->setSize(width, height);
    texture->setFormat(QOpenGLTexture::RGBAFormat);
    texture->allocateStorage();

    texture->bind();
    m_uploader_context->functions()->glCopyTexImage2D(GL_TEXTURE_2D,
                                                      0,
                                                      GL_RGBA,
                                                      src_x,
                                                      m_framebufferHeight - src_y,
                                                      width,
                                                      height,
                                                      0);
    texture->release();

    // "Specify the window coordinates of the lower left corner
    // of the rectangular region of pixels to be copied."
    m_copy_source->append({src_x,
                           m_framebufferHeight - src_y,
                           width,
                           height,
                           texture});

    // bottom left
    m_copy_data.append((float) dest_x);
    m_copy_data.append((float) (dest_y + height));
    m_copy_data.append(0.0f);
    m_copy_data.append(0.0f);
    // top left
    m_copy_data.append((float) dest_x);
    m_copy_data.append((float) dest_y);
    m_copy_data.append(0.0f);
    m_copy_data.append(1.0f);
    // bottom right
    m_copy_data.append((float) (dest_x + width));
    m_copy_data.append((float) (dest_y + height));
    m_copy_data.append(1.0f);
    m_copy_data.append(0.0f);
    // top right
    m_copy_data.append((float) (dest_x + width));
    m_copy_data.append((float) dest_y);
    m_copy_data.append(1.0f);
    m_copy_data.append(1.0f);

    m_copy_count++;
}

void Uploader::gotFillRect(const int x, const int y, const int width, const int height, const float red, const float green, const float blue)
{
    // |---|---|---|---|---|
    // | X | Y | R | G | B |
    // |---|---|---|---|---|
    // NOTE: -coordinates are transformed to normalized device coordinates in the vertex shader
    //       -coordinates and colors are extended to vec4 in the vertex shader
    //       -OpenGL ES 2.0 does not support integer values as input,
    //        thus we need to cast to float on the CPU side :(
    // bottom left
    m_fill_data.append((float) x);
    m_fill_data.append((float) (y + height));
    m_fill_data.append(red);
    m_fill_data.append(green);
    m_fill_data.append(blue);
    // top left
    m_fill_data.append((float) x);
    m_fill_data.append((float) y);
    m_fill_data.append(red);
    m_fill_data.append(green);
    m_fill_data.append(blue);
    // bottom right
    m_fill_data.append((float) (x + width));
    m_fill_data.append((float) (y + height));
    m_fill_data.append(red);
    m_fill_data.append(green);
    m_fill_data.append(blue);
    // top right
    m_fill_data.append((float) (x + width));
    m_fill_data.append((float) y);
    m_fill_data.append(red);
    m_fill_data.append(green);
    m_fill_data.append(blue);

    m_fill_count++;
}

void Uploader::gotBitmap(const QImage &image, const int x, const int y, const int width, const int height)
{
    m_draw_textures->append(new QOpenGLTexture(image.mirrored(), QOpenGLTexture::DontGenerateMipMaps));

    // bottom left
    m_draw_data.append((float) x);
    m_draw_data.append((float) (y + height));
    m_draw_data.append(0.0f);
    m_draw_data.append(0.0f);
    // top left
    m_draw_data.append((float) x);
    m_draw_data.append((float) y);
    m_draw_data.append(0.0f);
    m_draw_data.append(1.0f);
    // bottom right
    m_draw_data.append((float) (x + width));
    m_draw_data.append((float) (y + height));
    m_draw_data.append(1.0f);
    m_draw_data.append(0.0f);
    // top right
    m_draw_data.append((float) (x + width));
    m_draw_data.append((float) y);
    m_draw_data.append(1.0f);
    m_draw_data.append(1.0f);

    m_draw_count++;
}

void Uploader::finishedUpdate()
{
    if(m_fill_count > 0) {
        m_fill_buffer.bind();
        m_fill_buffer.allocate(m_fill_data.constData(), m_fill_data.size() * sizeof(float));
        m_fill_buffer.release();
    }

    if(m_copy_count > 0) {
        m_copy_buffer.bind();
        m_copy_buffer.allocate(m_copy_data.constData(), m_copy_data.size() * sizeof(float));
        m_copy_buffer.release();
    }

    if(m_draw_count > 0) {
        m_draw_buffer.bind();
        m_draw_buffer.allocate(m_draw_data.data(), m_draw_data.size() * sizeof(float));
        m_draw_buffer.release();
    }
    QOpenGLContext::currentContext()->functions()->glFlush();
    if(m_framebufferObject != NULL && m_framebufferObject->isBound())
        m_framebufferObject->release();
}

void Uploader::setFramebufferHeight(int framebufferHeight)
{
    m_framebufferHeight = framebufferHeight;
}

unsigned int Uploader::fill_count() const
{
    return m_fill_count;
}

unsigned int Uploader::copy_count() const
{
    return m_copy_count;
}

unsigned int Uploader::draw_count() const
{
    return m_draw_count;
}

QOpenGLBuffer Uploader::fill_buffer()
{
    return m_fill_buffer;
}

QSharedPointer<QVector<copy>> Uploader::copy_source()
{
    return m_copy_source;
}

QOpenGLBuffer Uploader::copy_buffer()
{
    return m_copy_buffer;
}

QOpenGLBuffer Uploader::draw_buffer()
{
    return m_draw_buffer;
}

QSharedPointer<QVector<QOpenGLTexture *>> Uploader::draw_textures()
{
    return m_draw_textures;
}

void Uploader::changeFramebufferObject(QOpenGLFramebufferObject *framebufferObject)
{
    m_framebufferObject = framebufferObject;
}
